use crate::util;
use clap::{Parser, Subcommand};
use der::{pem::PemLabel, referenced::OwnedToRef};
use image4::{
    der::{
        self,
        oid::{AssociatedOid, ObjectIdentifier},
        Document, Encode, EncodePem, Length, Writer,
    },
    property::Constraint,
    Tag,
};
use image4_pki::{DigestAlgo, PrivateKey, PublicKey, Signature, SigningKey, VerifyingKey};
use spki::{EncodePublicKey, SubjectPublicKeyInfoOwned};
use std::{collections::BTreeMap, io::Write, path::PathBuf, str::FromStr, time::Duration};
use x509_cert::{
    builder::{Builder, CertificateBuilder, Profile},
    ext::{AsExtension, Extension},
    name::Name,
    serial_number::SerialNumber,
    time::Validity,
    Certificate,
};

fn parse_path(s: &str) -> anyhow::Result<Vec<Tag>> {
    let s = s.trim();
    let mut tags = Vec::new();

    for comp in s.split('.') {
        tags.push(Tag::from_str(comp)?);
    }

    Ok(tags)
}

fn parse_constraint(s: &str) -> anyhow::Result<Constraint> {
    let s = s.trim();
    match s {
        "any" => Ok(Constraint::Any),
        "deny" => Ok(Constraint::Deny),
        "true" => Ok(Constraint::Boolean(true)),
        "false" => Ok(Constraint::Boolean(false)),
        "{}" => Ok(Constraint::Dict(Default::default())),
        "[]" => Ok(Constraint::Array(Default::default())),
        hex_s if hex_s.starts_with("0x") => hex::decode(&hex_s[2..])
            .map(Constraint::Data)
            .map_err(|e| anyhow::anyhow!("invalid data property format: {e}")),
        other => {
            if let Ok(i) = str::parse(other) {
                Ok(Constraint::Integer(i))
            } else {
                Err(anyhow::anyhow!(
                    "invalid constraint value \"{}\"",
                    other.escape_debug()
                ))
            }
        }
    }
}

fn evaluate_props(props: &Vec<CsrPropWithValue>) -> anyhow::Result<Constraint> {
    let mut constraint = BTreeMap::<Tag, Constraint>::new();

    for prop in props {
        let (&tag, path) = prop
            .path
            .split_last()
            .ok_or_else(|| anyhow::anyhow!("empty path"))?;

        let mut dest = &mut constraint;
        for comp in path {
            let Constraint::Dict(next_dict) = dest
                .get_mut(comp)
                .ok_or_else(|| anyhow::anyhow!("missing component \"{comp}\""))?
            else {
                return Err(anyhow::anyhow!("unexpected path component type"));
            };

            dest = next_dict;
        }

        dest.insert(tag, prop.value.clone());
    }

    Ok(Constraint::Dict(constraint))
}

#[derive(Clone, Debug)]
struct CsrPropWithValue {
    path: Vec<Tag>,
    value: Constraint,
}

impl FromStr for CsrPropWithValue {
    type Err = anyhow::Error;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let (path, value) = s
            .split_once('=')
            .ok_or_else(|| anyhow::anyhow!("invalid value spec"))?;

        let path = parse_path(path)?;
        let value = parse_constraint(value)?;

        Ok(Self { path, value })
    }
}

const APPLE_EXT_OID: ObjectIdentifier = ObjectIdentifier::new_unwrap("1.2.840.113635.100.6.1.15");

struct AppleExtRef<'a>(&'a Constraint);

impl AssociatedOid for AppleExtRef<'_> {
    const OID: ObjectIdentifier = APPLE_EXT_OID;
}

impl Encode for AppleExtRef<'_> {
    fn encoded_len(&self) -> der::Result<Length> {
        self.0.encoded_len()
    }

    fn encode(&self, encoder: &mut impl Writer) -> der::Result<()> {
        self.0.encode(encoder)
    }
}

impl AsExtension for AppleExtRef<'_> {
    fn critical(&self, _subject: &Name, _extensions: &[Extension]) -> bool {
        true
    }
}

#[derive(Parser, Debug)]
/// Creates a leaf X.509 certificate for signing Image4 manifests.
pub struct CreateArgs {
    /// Forces the tool to continue in cases when the user would be asked for confirmation.
    #[arg(long, short)]
    force: bool,

    /// The subject name to be included in the certificate.
    ///
    /// The value is expected to conform to the rules specified in RFC 4514.
    #[arg(long, short = 'S', value_name = "SUBJECT")]
    subject: Option<Name>,

    /// Specifies the path to a PEM-encoded X.509 certificate that should be the issuer of the
    /// generated certificate.
    #[arg(long, short = 'c', value_name = "PATH")]
    ca: PathBuf,

    /// Specifies the path to a PEM-encoded PKCS#1 or PKCS#8 private key for the CA.
    ///
    /// Only RSA and ECDSA NIST P-256 and P-384 PKCS#8 keys are supported.
    #[arg(long = "ca-private-key", short = 'k', value_name = "PATH")]
    ca_private_key: PathBuf,

    /// Specifies the path to the public key for the generated certificate.
    #[arg(long = "public-key", short = 'p', value_name = "PATH")]
    public_key: PathBuf,

    /// Digest algorithm to use for signing the certificate. Allowed values are "sha-1",
    /// "sha-256" and "sha-384".
    ///
    /// When this option is missing, the defaults will be used: SHA-384 for RSA keys, SHA-256 for
    /// the NIST P-256 curve and SHA-384 for the NIST P-384 curve.
    ///
    /// The actual supported algorithms depend on the type of key used for signing. RSA keys support
    /// any kind of digest, but ECDSA keys only support their default digests.
    #[arg(
        long = "digest",
        short = 'd',
        value_name = "ALGO",
        default_value = "sha384"
    )]
    digest_algo: Option<DigestAlgo>,

    /// Sets a property for the manifest constraints in the certificate.
    ///
    /// The format for the value is *path=value* where *path* is a sequence of 4CC tags split by
    /// dots and value is `any` for a wildcard constraint, `deny` for disallowing a tag, `true` for
    /// a boolean true, `false` for a boolean false, `{}` to create a dictionary, `[]` to create
    /// an array, a hexadecimal string prefixed with a `0x` for a byte buffer and a decimal
    /// integer for an unsigned integer.
    ///
    /// Signed integers, array indexing and strings are not supported.
    #[arg(long = "set", short = 's', value_name = "PATH=VALUE")]
    props: Vec<CsrPropWithValue>,

    /// Allows the private key used to sign the certificate to not match the public key in the
    /// specified CA certificate.
    #[arg(long = "allow-ca-key-mismatch")]
    allow_ca_key_mismatch: bool,

    /// A path where the newly created certificate will be written.
    ///
    /// When this option is missing, the file will be created adjacent to the input file. The name
    /// will be the same as that of the input file with the extension set to ".crt".
    #[arg(value_name = "OUTPUT")]
    output: PathBuf,
}

#[derive(Subcommand, Debug)]
#[command(name = "cert")]
/// Transforms X.509 certificates with support for custom Apple extensions for signing Image4
/// manifests.
pub enum CertCommand {
    Create(CreateArgs),
}

fn check_cert_private_key(cert: &Certificate, key: &PrivateKey) -> anyhow::Result<()> {
    let public_key = key.to_public_key();

    let spki = cert.tbs_certificate.subject_public_key_info.owned_to_ref();
    let cert_pub_key: PublicKey = spki.try_into()?;

    if public_key == cert_pub_key {
        Ok(())
    } else {
        Err(anyhow::anyhow!(concat!(
            "Signing key doesn't match public key of the CA. Specify --allow-ca-key-mismatch ",
            "argument if this is the intended behavior."
        )))
    }
}

pub fn do_create(args: &CreateArgs) -> anyhow::Result<()> {
    let constraint = evaluate_props(&args.props)?;
    let subject = args.subject.as_ref().unwrap().clone();

    let ca = Document::read_pem_file(&args.ca).and_then(|(label, ca)| {
        if label == Certificate::PEM_LABEL {
            Ok(ca)
        } else {
            Err(der::pem::Error::UnexpectedTypeLabel {
                expected: Certificate::PEM_LABEL,
            }
            .into())
        }
    })?;
    let ca: Certificate = ca.decode_msg()?;
    let key = PrivateKey::read_pem_file(&args.ca_private_key)?;

    if !args.allow_ca_key_mismatch {
        check_cert_private_key(&ca, &key)?;
    }

    let key = SigningKey::from_key_and_algo(key, args.digest_algo)?;
    let pub_key = PublicKey::read_pem_file(&args.public_key)?;
    let pub_key = VerifyingKey::from_key_and_algo(pub_key, args.digest_algo)?;
    let spki: SubjectPublicKeyInfoOwned = pub_key.to_public_key_der()?.decode_msg()?;

    info!(
        "Generating certificate for subject {subject} using key at {}.",
        args.public_key.display()
    );

    let mut builder = CertificateBuilder::new(
        Profile::Leaf {
            issuer: ca.tbs_certificate.issuer,
            enable_key_agreement: false,
            enable_key_encipherment: false,
        },
        SerialNumber::from(1337u64),
        Validity::from_now(Duration::from_secs(100000000))?,
        subject,
        spki,
        &key,
    )?;

    builder.add_extension(&AppleExtRef(&constraint))?;

    let cert = builder.build::<Signature>()?;
    let bytes = cert.to_pem(Default::default())?;

    info!(
        "Writing generated certificate to {}.",
        args.output.display()
    );

    let mut output = util::try_open_output(&args.output, args.force)?;
    output.write_all(bytes.as_bytes())?;

    Ok(())
}
