//! Based on public domain code by Haruhiko Okumura.
use adler32::RollingAdler32;
use std::mem;

const N: usize = 4096;
const F: usize = 18;
const THRESHOLD: u32 = 2;

pub trait Reader {
    fn read_byte(&mut self) -> Option<u8>;
}

impl<'a> Reader for &'a [u8] {
    fn read_byte(&mut self) -> Option<u8> {
        let (&b, tail) = self.split_first()?;
        *self = tail;
        Some(b)
    }
}

pub trait Writer {
    fn write_byte(&mut self, b: u8) -> bool;
}

impl Writer for &mut [u8] {
    fn write_byte(&mut self, b: u8) -> bool {
        if let Some((p, tail)) = mem::take(self).split_first_mut() {
            *p = b;
            *self = tail;
            true
        } else {
            false
        }
    }
}

impl Writer for Vec<u8> {
    fn write_byte(&mut self, b: u8) -> bool {
        self.push(b);
        true
    }
}

pub struct WriterWithAdler<W> {
    inner: W,
    adler: RollingAdler32,
}

impl<W> WriterWithAdler<W> {
    pub fn new(inner: W) -> Self {
        Self {
            inner,
            adler: RollingAdler32::new(),
        }
    }

    pub fn inner(&self) -> &W {
        &self.inner
    }

    pub fn hash(&self) -> u32 {
        self.adler.hash()
    }
}

impl<W: Writer> Writer for WriterWithAdler<W> {
    fn write_byte(&mut self, b: u8) -> bool {
        self.adler.update(b);
        self.inner.write_byte(b)
    }
}

pub fn decompress<R: Reader, W: Writer>(dst: &mut W, src: &mut R) {
    let mut text_buf = [b' '; N + F - 1];

    let mut r = N - F;
    let mut flags = 0;
    loop {
        flags >>= 1;

        if (flags & 0x100) == 0 {
            let Some(c) = src.read_byte() else { break };
            flags = (c as u32) | 0xFF00;
        }

        if flags & 1 != 0 {
            let Some(c) = src.read_byte() else { break };

            if !dst.write_byte(c) {
                break;
            }

            text_buf[r] = c;
            r = (r + 1) % N;
        } else {
            let Some(mut i) = src.read_byte().map(u32::from) else {
                break;
            };
            let Some(mut j) = src.read_byte().map(u32::from) else {
                break;
            };

            i |= (j & 0xF0) << 4;
            j = (j & 0x0F) + THRESHOLD;

            for k in 0..=j {
                let c = text_buf[((i + k) as usize) % N];

                if !dst.write_byte(c) {
                    break;
                }

                text_buf[r] = c;
                r = (r + 1) % N;
            }
        }
    }
}
