use crate::{comp::lzss::WriterWithAdler, util};
use image4::payload::{compr_info::COMPR_ALGO_LZFSE, ComprInfo};
use lzfse_rust::LzfseRingDecoder;
use std::{
    fmt::{Display, Formatter},
    io::Write,
    str::FromStr,
};

mod lzss;

#[derive(Copy, Clone, Eq, PartialEq, Debug)]
pub enum ComprAlgo {
    Lzfse,
    Lzss,
}

impl Display for ComprAlgo {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        f.write_str(match self {
            ComprAlgo::Lzfse => "LZFSE",
            ComprAlgo::Lzss => "LZSS",
        })
    }
}

impl FromStr for ComprAlgo {
    type Err = anyhow::Error;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let lower = s.to_ascii_lowercase();

        match lower.as_str() {
            "lzfse" => Ok(ComprAlgo::Lzfse),
            "lzss" => Ok(ComprAlgo::Lzss),
            "lzvn" => Err(anyhow::anyhow!("no LZVN yet sorry :(")),
            _ => Err(anyhow::anyhow!("unknown compression algorithm {s}")),
        }
    }
}

fn identify_algo_by_magic(bytes: &[u8]) -> Option<ComprAlgo> {
    if bytes.starts_with(b"bvx2") {
        Some(ComprAlgo::Lzfse)
    } else if bytes.starts_with(b"complzss") {
        Some(ComprAlgo::Lzss)
    } else {
        None
    }
}

fn decompress_lzss<W: Write>(data: &[u8], writer: &mut W) -> anyhow::Result<()> {
    if data.len() < 0x180 {
        return Err(anyhow::anyhow!("LZSS compressed buffer is too short."));
    }

    let exp_checksum: &[u8; 4] = (&data[8..12]).try_into()?;
    let exp_checksum = u32::from_be_bytes(*exp_checksum);
    let c_size: &[u8; 4] = (&data[16..20]).try_into()?;
    let c_size = u32::from_be_bytes(*c_size);
    let u_size: &[u8; 4] = (&data[12..16]).try_into()?;
    let u_size = u32::from_be_bytes(*u_size);

    let tail = &data[0x180..];
    if tail.len() < (c_size as usize) {
        return Err(anyhow::anyhow!(
            "Compressed size in the LZSS header is larger than the tail size ({} < {}).",
            c_size,
            tail.len()
        ));
    }

    let mut buf = WriterWithAdler::new(Vec::with_capacity(u_size as usize));
    let mut src = &tail[..(c_size as usize)];
    lzss::decompress(&mut buf, &mut src);

    if exp_checksum != buf.hash() {
        debug!("{exp_checksum:#x} != {:#x}", buf.hash());

        return Err(anyhow::anyhow!(
            "LZSS decompressor failed: output checksum is invalid"
        ));
    }

    writer.write_all(buf.inner())?;

    Ok(())
}

fn decompress_lzfse<W: Write>(
    data: &[u8],
    writer: &mut W,
    exp_size: Option<usize>,
) -> anyhow::Result<()> {
    // we know LZFSE compressed data ends with bvx$ so trim zeroes resulting from the
    // encryption alignment
    let mut ptr = util::trim_trailing_zeroes(data);

    debug!("Removed {} padding bytes.", data.len() - ptr.len());

    let mut decoder = LzfseRingDecoder::default();
    let (u, v) = decoder.decode(&mut ptr, writer)?;

    info!("Decompressed {u} bytes into {v} bytes.");

    if let Some(exp_size) = exp_size {
        if v != exp_size as u64 {
            warn!("Expected uncompressed size {exp_size}, got {v}.");
        }
    }

    Ok(())
}

pub fn maybe_decompress_payload<W: Write>(
    data: &[u8],
    compr_info: Option<&ComprInfo>,
    output_file: &mut W,
) -> anyhow::Result<()> {
    if let Some(compr_info) = compr_info {
        if compr_info.algo == COMPR_ALGO_LZFSE {
            decompress_lzfse(
                data,
                output_file,
                Some(compr_info.uncompressed_size.try_into()?),
            )
        } else {
            warn!(
                "Unsupported compression algorithm ID {}. Falling back to extracting raw data.",
                compr_info.algo
            );

            Ok(output_file.write_all(data)?)
        }
    } else if let Some(algo) = identify_algo_by_magic(data) {
        match algo {
            ComprAlgo::Lzfse => decompress_lzfse(data, output_file, None),
            ComprAlgo::Lzss => decompress_lzss(data, output_file),
        }
    } else {
        Ok(output_file.write_all(data)?)
    }
}
