use aes::{
    cipher::{
        block_padding::{NoPadding, UnpadError},
        BlockDecryptMut, Key, KeyIvInit,
    },
    Aes128, Aes192, Aes256,
};
use hex::FromHexError;
use std::{
    fmt::{self, Display, Formatter},
    str::FromStr,
};

#[derive(Clone, Debug)]
pub enum AesKey {
    Aes128(Key<cbc::Decryptor<Aes128>>),
    Aes192(Key<cbc::Decryptor<Aes192>>),
    Aes256(Key<cbc::Decryptor<Aes256>>),
}

impl AesKey {
    pub fn decrypt_with_iv_no_pad_vec(
        &self,
        iv: &AesIv,
        bytes: &[u8],
    ) -> Result<Vec<u8>, UnpadError> {
        match self {
            AesKey::Aes128(key) => {
                let cipher = cbc::Decryptor::<Aes128>::new(key, &iv.0.into());
                cipher.decrypt_padded_vec_mut::<NoPadding>(bytes)
            }
            AesKey::Aes192(key) => {
                let cipher = cbc::Decryptor::<Aes192>::new(key, &iv.0.into());
                cipher.decrypt_padded_vec_mut::<NoPadding>(bytes)
            }
            AesKey::Aes256(key) => {
                let cipher = cbc::Decryptor::<Aes256>::new(key, &iv.0.into());
                cipher.decrypt_padded_vec_mut::<NoPadding>(bytes)
            }
        }
    }
}

impl FromStr for AesKey {
    type Err = FromHexError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let k = hex::decode(s)?;
        let k = k.as_slice();

        if let Ok(k) = TryInto::<&[u8; 16]>::try_into(k) {
            Ok(AesKey::Aes128(From::from(*k)))
        } else if let Ok(k) = TryInto::<&[u8; 24]>::try_into(k) {
            Ok(AesKey::Aes192(From::from(*k)))
        } else if let Ok(k) = TryInto::<&[u8; 32]>::try_into(k) {
            Ok(AesKey::Aes256(From::from(*k)))
        } else {
            Err(FromHexError::InvalidStringLength)
        }
    }
}

impl Display for AesKey {
    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
        match self {
            AesKey::Aes128(key) => write!(f, "AES-128 key {}", hex::encode(key)),
            AesKey::Aes192(key) => write!(f, "AES-192 key {}", hex::encode(key)),
            AesKey::Aes256(key) => write!(f, "AES-256 key {}", hex::encode(key)),
        }
    }
}

#[derive(Clone, Debug)]
pub struct AesIv([u8; 16]);

impl Display for AesIv {
    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
        f.write_str(&hex::encode(self.0))
    }
}

impl FromStr for AesIv {
    type Err = FromHexError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let mut iv = AesIv([0u8; 16]);
        hex::decode_to_slice(s, &mut iv.0)?;
        Ok(iv)
    }
}
