#[macro_use]
extern crate log;

mod cert;
mod comp;
mod enc;
mod manifest;
mod payload;
mod util;

use crate::{
    cert::CertCommand,
    enc::{AesIv, AesKey},
    manifest::ManifestCommand,
    payload::{PayloadCommand, PayloadCreationOptions},
    util::ImageOrPayloadRef,
};
use clap::{command, FromArgMatches, Parser, Subcommand};
use der::{referenced::OwnedToRef, Decode};
use image4::{der::Encode, ImageRef, RestoreInfo, Tag};
use log::{Level, LevelFilter};
use std::{
    borrow::Cow,
    fmt::Debug,
    fs::File,
    io::{Read, Write},
    path::PathBuf,
    process::ExitCode,
};

const DEFAULT_BUILDER_STRING: &str = concat!("image4_util-", env!("CARGO_PKG_VERSION"));

#[derive(Parser, Debug)]
/// Creates a full Image4 file from raw payload contents or an Image4 file.
struct CreateArgs {
    /// A 4CC tag for the payload of the output file.
    ///
    /// When this option is present, the input file is treated as raw payload contents rather than
    /// an Image4 payload file (IM4P).
    #[arg(long, short)]
    tag: Option<Tag>,

    #[command(flatten)]
    common: PayloadCreationOptions,

    /// Adds an Image4 manifest to the output file.
    #[arg(long, short)]
    manifest: Option<PathBuf>,

    /// Adds a restore info blob (IM4R) to the generate Image4 file containing the specified nonce.
    #[arg(long = "boot-nonce")]
    boot_nonce: Option<u64>,

    /// Forces the tool to continue in cases when the user would be asked for confirmation.
    #[arg(long, short)]
    force: bool,

    /// A path to the input file containing the payload to be wrapped into the Image4 format.
    ///
    /// When the -t/--tag argument is present the file is treated as raw data. The payload for the
    /// Image4 file will be generated with the specified tag. When the tag is not present, the file
    /// is expected to be an Image4 payload file (IM4P).
    #[arg(value_name = "INPUT")]
    input: PathBuf,

    /// A path where the created Image4 file should be written. When missing, the output file will
    /// be created in the same directory as the input file.
    #[arg(value_name = "OUTPUT")]
    output: Option<PathBuf>,
}

#[derive(Parser, Debug)]
/// Extracts either a full Image4 image (IMG4) or an Image4 payload (IM4P) file.
struct ExtractArgs {
    /// Key for decryption. Requires -I/--iv argument to be present.
    #[arg(long, short = 'K', requires = "iv")]
    key: Option<AesKey>,

    /// Initialization vector for decryption. Requires -K/--key argument to be present.
    #[arg(long, short = 'I', requires = "key")]
    iv: Option<AesIv>,

    /// Forces the tool to skip decompression entirely.
    ///
    /// This will print a warning if the file contains decompression info. Can be useful in case
    /// compression info contains an unknown algorithm.
    #[arg(long = "no-decompress")]
    no_decompress: bool,

    /// The tool will not try to detect if the file is compressed in case there is no compression
    /// info in the payload.
    #[arg(long = "no-auto-decompress")]
    no_auto_decompress: bool,

    /// Parses payload properties and writes them to an XML Property List file.
    #[arg(long = "payload-props")]
    props: Option<PathBuf>,

    /// Forces the tool to continue in cases when the user would be asked for confirmation.
    #[arg(long, short)]
    force: bool,

    /// A path to the file to be extracted.
    #[arg(value_name = "INPUT")]
    input: PathBuf,

    /// Destination where the file should be written.
    ///
    /// If missing, a file will be created in the same directory as the input file.
    #[arg(value_name = "OUTPUT")]
    output: Option<PathBuf>,
}

#[derive(Parser, Debug)]
/// Splits a full Image4 image (IMG4) file into its components.
struct SplitArgs {
    /// Forces the tool to continue in cases when the user would be asked for confirmation.
    #[arg(long, short)]
    force: bool,

    /// Prints the boot nonce contained in the image's restore info.
    ///
    /// When this option is present and there is no restore info in the provided image, an error
    /// will occur.
    #[arg(long = "print-boot-nonce", short = 'r')]
    print_boot_nonce: bool,

    /// A path to the full Image4 file to be split.
    #[arg(value_name = "INPUT")]
    input: PathBuf,

    /// A path where the payload extracted from the image should be written.
    ///
    /// If missing, the payload will not be extracted.
    #[arg(long, short, value_name = "PATH")]
    payload: Option<PathBuf>,

    /// A path where the manifest extracted from the image should be written.
    ///
    /// If missing, the manifest will not be extracted. When this option is present and there is no
    /// manifest in the provided image, an error will occur.
    #[arg(long, short, value_name = "PATH")]
    manifest: Option<PathBuf>,
}

#[derive(Parser, Debug)]
enum TopLevelCommands {
    Create(CreateArgs),
    Extract(ExtractArgs),
    Split(SplitArgs),
    #[command(subcommand)]
    Payload(PayloadCommand),
    #[command(subcommand)]
    Manifest(ManifestCommand),
    #[command(subcommand)]
    Cert(CertCommand),
}

fn do_extract(args: &ExtractArgs) -> anyhow::Result<()> {
    let mut input_file = File::open(&args.input)?;

    let output_path = util::construct_output_path(
        args.output.as_deref(),
        &args.input,
        "bin",
        &["img4", "im4p"],
    );

    info!(
        "Extracting file at \'{}\' to \'{}\'.",
        args.input.display(),
        output_path.display()
    );

    let mut input_contents = Vec::new();
    input_file.read_to_end(&mut input_contents)?;

    let input_contents = match util::parse_image_or_payload(&input_contents) {
        Ok(c) => c,
        Err(e) => {
            return Err(anyhow::Error::msg(format!(
                "couldn't parse image at \'{}\': {}",
                args.input.display(),
                e
            )));
        }
    };
    let payload = match &input_contents {
        ImageOrPayloadRef::Image(image) => image.payload(),
        ImageOrPayloadRef::Payload(payload) => payload,
    };

    info!("4CC tag: {}", payload.tag());
    info!("Builder string: {}", payload.builder_string());

    let data = match (&args.key, &args.iv) {
        (Some(key), Some(iv)) => {
            if payload.keybags().is_none() {
                warn!("Decrypting an image that is not encrypted.");
            }

            debug!("Decrypting using {}, IV {}.", key, iv);

            Cow::Owned(key.decrypt_with_iv_no_pad_vec(iv, payload.data())?)
        }
        (Some(_), None) | (None, Some(_)) => unreachable!(),
        (None, None) => {
            if payload.keybags().is_some() {
                warn!("The payload is encrypted, no decryption keys were provided.")
            }

            Cow::Borrowed(payload.data())
        }
    };

    if let Some(props_path) = &args.props {
        if let Some(props) = payload.props() {
            info!(
                "Writing payload properties to file at {}.",
                props_path.display()
            );

            let props = props.decode_body()?;
            plist::to_file_xml(props_path, &props)?;
        } else {
            warn!("Requested saving payload properties but the payload lacks properties.");
        }
    }

    let mut output_file = util::try_open_output(&output_path, args.force)?;

    if !args.no_decompress && (!args.no_auto_decompress || payload.compr_info().is_some()) {
        comp::maybe_decompress_payload(&data, payload.compr_info(), &mut output_file)
    } else {
        if payload.compr_info().is_some() {
            warn!("Not decompressing even though compression info is present.");
        }

        Ok(output_file.write_all(&data)?)
    }
}

fn do_create(args: &CreateArgs) -> anyhow::Result<()> {
    let mut input_file = File::open(&args.input)?;
    let mut input_data = Vec::new();
    input_file.read_to_end(&mut input_data)?;

    let output_path = util::construct_output_path(
        args.output.as_deref(),
        &args.input,
        "img4",
        &["im4p", "bin"],
    );

    info!(
        "Creating IMG4 file at {} from {}.",
        output_path.display(),
        args.input.display()
    );

    let mut compr_info = None;

    let mut image = if let Some(tag) = args.tag {
        debug!("Treating input as raw data.");

        compr_info = util::maybe_compress_payload(args.common.compr_algo, &mut input_data)?;

        let builder_string = args
            .common
            .builder_string
            .as_deref()
            .unwrap_or(DEFAULT_BUILDER_STRING);
        ImageRef::new(tag, builder_string, &input_data).map_err(Into::into)
    } else {
        debug!("Treating input as a payload file.");

        if let Some(s) = &args.common.builder_string {
            warn!("Input file is a payload file. Specified builder string '{s}' will be ignored.");
        }

        ImageRef::from_payload_bytes(&input_data).map_err(|e| {
            anyhow::anyhow!(
                "Failed to parse payload file at '{}': {e}",
                args.input.display()
            )
        })
    }?;

    info!("4CC tag: {}", image.payload().builder_string());
    info!("Builder string: {}", image.payload().builder_string());

    let mut manifest_data = Vec::new();
    if let Some(path) = args.manifest.as_deref() {
        info!("Loading manifest file at {}.", path.display());

        let mut manifest_file = File::open(path)?;
        manifest_file.read_to_end(&mut manifest_data)?;

        image.parse_manifest(&manifest_data)?;
    }

    if let Some(boot_nonce) = args.boot_nonce {
        info!("Adding a restore info blob with nonce {boot_nonce:#x}.");

        image.set_restore_info(Some(RestoreInfo {
            nonce_seed: boot_nonce,
        }));
    }

    if let Some(compr_info) = &args.common.raw_compr_info {
        info!(
            "Adding raw compression info (algo is {}, uncompressed size is {}).",
            compr_info.algo, compr_info.uncompressed_size
        );

        image.payload_mut().set_compr_info(Some(compr_info.clone()));
    } else if let Some(compr_info) = compr_info {
        image.payload_mut().set_compr_info(Some(compr_info));
    }

    let props = util::maybe_load_payload_props(args.common.props.as_ref())?;
    if let Some(props) = &props {
        image.payload_mut().set_props(Some(props.owned_to_ref()));
    }

    let mut output_file = util::try_open_output(&output_path, args.force)?;
    let mut data = Vec::new();
    image.encode_to_vec(&mut data)?;
    output_file.write_all(&data)?;

    Ok(())
}

fn do_split(args: &SplitArgs) -> anyhow::Result<()> {
    info!(
        "Splitting full Image4 file at path {}.",
        args.input.display()
    );

    let mut input_file = File::open(&args.input)?;
    let mut input_data = Vec::new();
    input_file.read_to_end(&mut input_data)?;

    let image = ImageRef::from_der(&input_data)?;

    if args.manifest.is_some() || args.payload.is_some() || args.print_boot_nonce {
        let manifest = util::ensure_left_some_and_right_some_or_else(
            args.manifest.as_deref(),
            image.manifest(),
            || anyhow::anyhow!("input file is missing a manifest"),
        )?;

        let restore_info = util::ensure_left_some_and_right_some_or_else(
            args.print_boot_nonce.then_some(()),
            image.restore_info(),
            || anyhow::anyhow!("input file is missing restore info"),
        )?;

        if let Some(payload_path) = args.payload.as_ref() {
            info!("Writing payload to {}.", payload_path.display());

            let mut output_file = util::try_open_output(payload_path, args.force)?;
            let mut v = Vec::new();
            image.payload().encode_to_vec(&mut v)?;
            output_file.write_all(&v)?
        }

        if let Some((manifest_path, manifest)) = manifest {
            info!("Writing manifest to {}.", manifest_path.display());

            let mut output_file = util::try_open_output(manifest_path, args.force)?;

            let mut v = Vec::new();
            manifest.encode_to_vec(&mut v)?;
            output_file.write_all(&v)?;
        }

        if let Some((_, restore_info)) = restore_info {
            println!("Boot nonce: {:#x}", restore_info.nonce_seed);
        }
    } else {
        warn!("None of the input Image4 file components were requested during split.")
    }

    Ok(())
}

fn level_to_prefix(l: Level) -> &'static str {
    match l {
        Level::Error => "\x1b[31m[-]\x1b[0m",
        Level::Warn => "\x1b[33m[!]\x1b[0m",
        Level::Info => "[*]",
        Level::Debug => "[d]",
        Level::Trace => "[t]",
    }
}

fn main() -> ExitCode {
    env_logger::Builder::default()
        .format(|f, r| writeln!(f, "{} {}", level_to_prefix(r.level()), r.args()))
        .filter_level(LevelFilter::Warn)
        .parse_env("IMAGE4_LOG")
        .init();

    let command = command!().subcommand_required(true);
    let command = TopLevelCommands::augment_subcommands(command);
    let matches = command.get_matches();

    let result = TopLevelCommands::from_arg_matches(&matches)
        .map_err(anyhow::Error::from)
        .and_then(|commands| match commands {
            TopLevelCommands::Create(args) => do_create(&args),
            TopLevelCommands::Extract(args) => do_extract(&args),
            TopLevelCommands::Split(args) => do_split(&args),
            TopLevelCommands::Payload(commands) => match commands {
                PayloadCommand::Create(args) => payload::do_create(&args),
            },
            TopLevelCommands::Manifest(commands) => match commands {
                ManifestCommand::Encode(args) => manifest::do_encode(&args),
                ManifestCommand::Sign(args) => manifest::do_sign(&args),
                ManifestCommand::Print(args) => manifest::do_print(&args),
                ManifestCommand::Parse(args) => manifest::do_parse(&args),
            },
            TopLevelCommands::Cert(commands) => match commands {
                CertCommand::Create(args) => cert::do_create(&args),
            },
        });

    match result {
        Ok(()) => {
            info!("Success.");
            ExitCode::SUCCESS
        }
        Err(e) => {
            error!("{}", e);
            ExitCode::FAILURE
        }
    }
}
