use crate::util;
use clap::{Parser, Subcommand};
use image4::{
    der::{Decode, Encode},
    manifest::{AnyManifestRef, UnsignedManifest},
    property::Value,
    Manifest, Tag,
};
use image4_pki::{DigestAlgo, PrivateKey, SigningKey};
use std::{
    collections::BTreeMap,
    fs::File,
    io::{Read, Write},
    path::PathBuf,
};

#[derive(Parser, Debug)]
/// Encodes a Property List describing manifest's structure into an Image4 manifest.
pub struct EncodeArgs {
    /// Forces the tool to continue in cases when the user would be asked for confirmation.
    #[arg(long, short)]
    force: bool,

    /// The digest algorithm to be used when signing the manifest. Allowed values are "sha-1",
    /// "sha-256" and "sha-384".
    ///
    /// When this option is missing, the defaults will be used: SHA-384 for RSA keys, SHA-256 for
    /// the NIST P-256 curve and SHA-384 for the NIST P-384 curve.
    ///
    /// The actual supported algorithms depend on the type of key used for signing. RSA keys support
    /// any kind of digest, but ECDSA keys only support their default digests.
    #[arg(
        long = "digest",
        short = 'd',
        requires = "cert_chain",
        value_name = "ALGO"
    )]
    digest_algo: Option<DigestAlgo>,

    /// Specifies the path to a PEM-encoded PKCS#1 or PKCS#8 private key to be used to sign the
    /// manifest.
    ///
    /// Only RSA and ECDSA NIST P-256 and P-384 PKCS#8 keys are supported.
    #[arg(
        long = "private-key",
        short = 'k',
        requires = "cert_chain",
        value_name = "PATH"
    )]
    private_key: Option<PathBuf>,

    /// Specifies the path to a PEM-encoded certificate chain to be used to sign the manifest.
    #[arg(
        long = "cert-chain",
        short = 'c',
        requires = "private_key",
        value_name = "PATH"
    )]
    cert_chain: Option<PathBuf>,

    /// A path to the Property List describing the manifest's body.
    #[arg(value_name = "INPUT")]
    input: PathBuf,

    /// A path where the generated manifest should be written.
    #[arg(value_name = "OUTPUT")]
    output: Option<PathBuf>,
}

#[derive(Parser, Debug)]
/// Signs an unsigned Image4 manifest.
pub struct SignArgs {
    /// Forces the tool to continue in cases when the user would be asked for confirmation.
    #[arg(short, long)]
    force: bool,

    /// The digest algorithm to be used when signing the manifest. Allowed values are "sha-1",
    /// "sha-256" and "sha-384".
    ///
    /// When this option is missing, the defaults will be used: SHA-384 for RSA keys, SHA-256 for
    /// the NIST P-256 curve and SHA-384 for the NIST P-384 curve.
    ///
    /// The actual supported algorithms depend on the type of key used for signing. RSA keys support
    /// any kind of digest, but ECDSA keys only support their default digests.
    #[arg(
        long = "digest",
        short = 'd',
        requires = "cert_chain",
        value_name = "ALGO"
    )]
    digest_algo: Option<DigestAlgo>,

    /// Specifies the path to a PEM-encoded PKCS#1 or PKCS#8 private key to be used to sign the
    /// manifest.
    ///
    /// Only RSA and ECDSA NIST P-256 and P-384 PKCS#8 keys are supported.
    #[arg(short = 'k', long = "private-key", value_name = "PATH")]
    private_key: PathBuf,

    /// Specifies the path to a PEM-encoded certificate chain to be used to sign the manifest.
    #[arg(long = "cert-chain", short = 'c', value_name = "PATH")]
    cert_chain: PathBuf,

    /// A path to the manifest to be signed.
    ///
    /// This can be both an unsigned or a signed manifest. In the latter case the signature will be
    /// stripped and a new one will be generated.
    #[arg(value_name = "INPUT")]
    input: PathBuf,

    /// A path where the signed manifest should be written.
    #[arg(value_name = "OUTPUT")]
    output: Option<PathBuf>,
}

#[derive(Parser, Debug)]
/// Converts the contents of an Image4 manifest body into the Property List format.
pub struct ParseArgs {
    /// Forces the tool to continue in cases when the user would be asked for confirmation.
    #[arg(long, short)]
    force: bool,

    /// Write the output in the binary Property List format instead of XML.
    #[arg(long, short)]
    binary: bool,

    /// A path to the manifest file to be printed.
    #[arg(value_name = "INPUT")]
    input: PathBuf,

    /// A path where the parsed output should be written.
    #[arg(value_name = "OUTPUT")]
    output: Option<PathBuf>,
}

#[derive(Parser, Debug)]
/// Prints the contents of an Image4 manifest body.
pub struct PrintArgs {
    /// A path to the manifest file to be printed.
    #[arg(value_name = "INPUT")]
    input: PathBuf,
}

#[derive(Subcommand, Debug)]
#[command(name = "manifest")]
/// Transforms Image4 manifests (IM4M files).
pub enum ManifestCommand {
    Encode(EncodeArgs),
    Sign(SignArgs),
    Print(PrintArgs),
    Parse(ParseArgs),
}

pub fn do_encode(args: &EncodeArgs) -> anyhow::Result<()> {
    let output =
        util::construct_output_path(args.output.as_deref(), &args.input, "im4m", &["plist"]);

    info!(
        "Generating a manifest at {} using the description at {}.",
        args.input.display(),
        output.display()
    );

    let body: BTreeMap<Tag, Value> = plist::from_file(&args.input)?;

    let mut data = Vec::new();
    if let (Some(private_key_path), Some(cert_chain_path)) =
        (args.private_key.as_ref(), args.cert_chain.as_ref())
    {
        info!(
            "Signing generated manifest using key at {}, certificate chain at {}",
            private_key_path.display(),
            cert_chain_path.display()
        );

        let key = PrivateKey::read_pem_file(private_key_path)?;
        let key = SigningKey::from_key_and_algo(key, args.digest_algo)?;
        let cert_chain = util::load_pem_cert_chain(cert_chain_path)?;

        let manifest = Manifest::encode_and_sign(&body, &key, cert_chain)?;
        manifest.encode_to_vec(&mut data)
    } else {
        let manifest = UnsignedManifest::encode_from(&body)?;

        manifest.encode_to_vec(&mut data)
    }?;

    util::try_open_output(&output, args.force)?.write_all(&data)?;

    Ok(())
}

pub fn do_sign(args: &SignArgs) -> anyhow::Result<()> {
    let output = util::construct_output_path(
        args.output.as_deref(),
        &args.input,
        "im4m",
        &["der", "im4m"],
    );

    info!(
        "Signing a manifest at {}, saving as {}.",
        args.input.display(),
        output.display()
    );

    let mut manifest_data = Vec::new();
    File::open(&args.input)?.read_to_end(&mut manifest_data)?;
    let manifest_to_be_signed = AnyManifestRef::from_der(&manifest_data)?;

    info!(
        "{} at {}, saving at {}.",
        match manifest_to_be_signed {
            AnyManifestRef::Signed(_) => "Resigning a signed manifest",
            AnyManifestRef::Unsigned(_) => "Signing an unsigned manifest",
        },
        args.input.display(),
        output.display()
    );

    let key = PrivateKey::read_pem_file(&args.private_key)?;
    let key = SigningKey::from_key_and_algo(key, args.digest_algo)?;
    let cert_chain = util::load_pem_cert_chain(&args.cert_chain)?;

    let mut manifest_data = Vec::new();
    let manifest = manifest_to_be_signed.as_ref().sign(&key, cert_chain)?;
    manifest.encode_to_vec(&mut manifest_data)?;

    util::try_open_output(&output, args.force)?.write_all(&manifest_data)?;

    Ok(())
}

pub fn do_print(args: &PrintArgs) -> anyhow::Result<()> {
    let mut manifest_file = File::open(&args.input)?;
    let mut manifest = Vec::new();
    manifest_file.read_to_end(&mut manifest)?;

    let manifest = AnyManifestRef::from_der(&manifest)?;
    let body = manifest.decode_body()?;

    util::print_dict(&body, false);
    println!();

    Ok(())
}

pub fn do_parse(args: &ParseArgs) -> anyhow::Result<()> {
    info!("Loading manifest file at {}.", args.input.display());

    let mut manifest_file = File::open(&args.input)?;
    let mut manifest = Vec::new();
    manifest_file.read_to_end(&mut manifest)?;

    let manifest = AnyManifestRef::from_der(&manifest)?;
    let body = manifest.decode_body().map(Value::Dict)?;

    let output_path = util::construct_output_path(
        args.output.as_deref(),
        &args.input,
        "plist",
        &["im4m", "der"],
    );

    info!(
        "Writing parsed manifest body to {} as an XML Plist.",
        output_path.display()
    );

    Ok(if args.binary {
        plist::to_file_binary(output_path, &body)
    } else {
        plist::to_file_xml(output_path, &body)
    }?)
}
