use crate::comp::ComprAlgo;
use image4::{
    der,
    der::{asn1::Ia5StringRef, ErrorKind, Header, Length, Reader, SliceReader},
    manifest::CertChain,
    payload::{compr_info::COMPR_ALGO_LZFSE, ComprInfo, PayloadProps},
    property::Value,
    ImageRef, PayloadRef, Tag,
};
use lzfse_rust::LzfseEncoder;
use std::{
    borrow::Cow,
    collections::BTreeMap,
    ffi::OsString,
    fs::{File, OpenOptions},
    io,
    io::Read,
    mem,
    ops::Add,
    path::Path,
    str::FromStr,
};
use x509_cert::Certificate;

pub fn try_open_output(output_path: &Path, overwrite: bool) -> anyhow::Result<File> {
    OpenOptions::new()
        .write(true)
        .create_new(!overwrite)
        .create(true)
        .truncate(true)
        .open(output_path)
        .or_else::<anyhow::Error, _>(|e| match e.kind() {
            io::ErrorKind::AlreadyExists => {
                let replace = dialoguer::Confirm::new()
                    .with_prompt(format!(
                        "[?] File at {} exists. Do you want to replace it?",
                        output_path.display()
                    ))
                    .interact()?;

                if replace {
                    Ok(OpenOptions::new()
                        .write(true)
                        .truncate(true)
                        .open(output_path)?)
                } else {
                    Err(e.into())
                }
            }
            _ => Err(e.into()),
        })
}

#[derive(Copy, Clone, Eq, PartialEq, Debug)]
pub enum FileType {
    Full,
    Payload,
    Manifest,
}

pub fn parse_with<'a, T, F>(bytes: &'a [u8], f: F) -> anyhow::Result<T>
where
    F: FnOnce(FileType, &mut SliceReader<'a>) -> anyhow::Result<T>,
{
    let mut reader = SliceReader::new(bytes)?;

    let header = reader.decode::<Header>()?;
    if header.tag != der::Tag::Sequence {
        return Err(ErrorKind::TagUnexpected {
            expected: Some(der::Tag::Sequence),
            actual: header.tag,
        }
        .at(Length::ZERO)
        .into());
    } else if header.length < reader.remaining_len() {
        return Err(reader
            .error(ErrorKind::TrailingData {
                decoded: header.length.add(reader.position())?,
                remaining: reader.remaining_len(),
            })
            .into());
    } else if header.length > reader.remaining_len() {
        return Err(reader
            .error(ErrorKind::Incomplete {
                expected_len: header.length,
                actual_len: reader.remaining_len(),
            })
            .into());
    }

    let magic = reader.decode::<Ia5StringRef<'_>>()?;
    let kind = match magic.as_str() {
        "IMG4" => FileType::Full,
        "IM4P" => FileType::Payload,
        "IM4M" => FileType::Manifest,
        other => {
            return Err(anyhow::Error::msg(format!(
                "unsupported magic {}",
                other.escape_debug()
            )))
        }
    };

    f(kind, &mut reader)
}

pub enum ImageOrPayloadRef<'a> {
    Image(ImageRef<'a>),
    Payload(PayloadRef<'a>),
}

pub(crate) fn parse_image_or_payload(bytes: &[u8]) -> anyhow::Result<ImageOrPayloadRef<'_>> {
    parse_with(bytes, |kind, reader| match kind {
        FileType::Full => Ok(ImageRef::decode_after_magic(reader).map(ImageOrPayloadRef::Image)?),
        FileType::Payload => {
            Ok(PayloadRef::decode_after_magic(reader).map(ImageOrPayloadRef::Payload)?)
        }
        FileType::Manifest => Err(anyhow::Error::msg(
            "expected either a full image or a payload, got a manifest",
        )),
    })
}

pub fn construct_output_path<'a>(
    output_path: Option<&'a Path>,
    input_path: &Path,
    ext: &str,
    replaced_exts: &[&str],
) -> Cow<'a, Path> {
    if let Some(path) = output_path {
        Cow::from(path)
    } else {
        Cow::from(if let Some(input_ext) = input_path.extension() {
            if replaced_exts
                .iter()
                .any(|&replaced_ext| replaced_ext == input_ext)
            {
                input_path.with_extension(ext)
            } else {
                let mut new_ext = OsString::with_capacity(input_ext.len() + 1 + ext.len());
                new_ext.push(input_ext);
                new_ext.push(".");
                new_ext.push(ext);
                input_path.with_extension(new_ext)
            }
        } else {
            input_path.with_extension(ext)
        })
    }
}

pub fn trim_trailing_zeroes(data: &[u8]) -> &[u8] {
    let mut ptr = data;
    let mut left = ptr.len();
    while left > 0 {
        left -= 1;

        if ptr[left] == 0 {
            ptr = &ptr[..left];
        } else {
            break;
        }
    }
    ptr
}

fn print_dict_internal(dict: &BTreeMap<Tag, Value>, flat: bool, padding: usize) {
    print!("{{");

    let inner_padding = padding + 2;
    let mut first = true;
    for (t, v) in dict {
        if !first {
            print!(",");
        } else {
            first = false;
        }

        if flat {
            print!(" ");
        } else {
            println!();
            for _ in 0..inner_padding {
                print!(" ");
            }
        }

        print!("'{}' = ", t);

        print_internal(v, flat, inner_padding);
    }

    if flat {
        print!(" }}");
    } else {
        println!();
        for _ in 0..padding {
            print!(" ");
        }
        print!("}}");
    }
}

fn print_internal(value: &Value, flat: bool, padding: usize) {
    match value {
        Value::Boolean(v) => print!("{:?}", v),
        Value::Integer(v) => print!("{}", v),
        Value::Data(v) => print!("<{}>", hex::encode(v)),
        Value::Dict(dict) => {
            print_dict_internal(dict, flat, padding);
        }
        Value::Array(array) => {
            print!("[");

            let mut first = true;
            for v in array {
                if !first {
                    print!(",");
                } else {
                    first = false;
                }

                if flat {
                    print!(" ");
                } else {
                    println!();
                    for _ in 0..padding {
                        print!(" ");
                    }
                }

                print_internal(v, flat, padding + 2);
            }

            if flat {
                print!(" ]");
            } else {
                println!();
                print!("]");
            }
        }
    }
}

pub fn print_dict(dict: &BTreeMap<Tag, Value>, flat: bool) {
    print_dict_internal(dict, flat, 0);
}

pub fn parse_raw_compr_info(s: &str) -> anyhow::Result<ComprInfo> {
    let mut parts = s.split(',');

    let algo = parts
        .next()
        .ok_or_else(|| anyhow::anyhow!("empty raw compression info string"))?;
    let algo = u32::from_str(algo)?;

    let uncompressed_size = parts.next().ok_or_else(|| {
        anyhow::anyhow!("missing uncompressed size in raw compression info string")
    })?;
    let uncompressed_size = u32::from_str(uncompressed_size)?;

    if parts.next().is_none() {
        Ok(ComprInfo {
            algo,
            uncompressed_size,
        })
    } else {
        Err(anyhow::anyhow!(
            "unexpected component in raw compression info string"
        ))
    }
}

pub fn maybe_compress_payload(
    algo: Option<ComprAlgo>,
    input_data: &mut Vec<u8>,
) -> anyhow::Result<Option<ComprInfo>> {
    if let Some(compr_algo) = algo {
        if compr_algo != ComprAlgo::Lzfse {
            return Err(anyhow::anyhow!(
                "Only LZFSE compression is supported at the moment."
            ));
        }

        info!("Compressing payload data using LZFSE.");

        let to_compress = mem::take(input_data);
        let mut encoder = LzfseEncoder::default();
        encoder.encode_bytes(&to_compress, input_data)?;

        info!(
            "Compressed payload from {} to {} bytes.",
            to_compress.len(),
            input_data.len()
        );

        Ok(Some(ComprInfo {
            algo: COMPR_ALGO_LZFSE,
            uncompressed_size: to_compress.len().try_into()?,
        }))
    } else {
        Ok(None)
    }
}

pub fn load_pem_cert_chain(path: impl AsRef<Path>) -> anyhow::Result<CertChain> {
    let mut v = Vec::new();
    File::open(path)?.read_to_end(&mut v)?;

    let certs = Certificate::load_pem_chain(&v)?;
    Ok(CertChain::from_certs(&certs)?)
}

pub fn ensure_left_some_and_right_some_or_else<L, R, F>(
    left: Option<L>,
    right: Option<R>,
    err: F,
) -> anyhow::Result<Option<(L, R)>>
where
    F: FnOnce() -> anyhow::Error,
{
    match (left, right) {
        (Some(path), Some(value)) => Ok(Some((path, value))),
        (None, _) => Ok(None),
        (Some(_), None) => Err(err()),
    }
}

pub fn maybe_load_payload_props<P: AsRef<Path>>(
    path: Option<P>,
) -> anyhow::Result<Option<PayloadProps>> {
    path.map(|path| {
        let path = path.as_ref();

        info!("Adding payload properties from file at {}.", path.display());

        let props_dict: BTreeMap<Tag, Value> = plist::from_file(path)?;
        Ok(PayloadProps::encode_from(&props_dict)?)
    })
    .transpose()
}
