//! A `no_std`-friendly library for parsing and generation of Image4 images written in pure Rust.
//!
//! This library is currently in **EXPERIMENTAL** state. It's very likely that the API changes in
//! the future.
//!
//! # Usage
//!
//! This library is based on the [`der`] crate. Decoding and encoding is based on the [`Decode`] and
//! [`Encode`] traits.
//!
//! An Image4 file may be decoded as follows:
//!
//! ```
//! use image4::{der::Decode, ImageRef, Tag};
//!
//! # fn test() -> der::Result<()> {
//! let bytes = include_bytes!("../tests/data/test.img4");
//! let image = ImageRef::from_der(bytes)?;
//!
//! assert_eq!(image.payload().tag(), Tag::from_bytes(*b"TEST"));
//! # Ok(())
//! # }
//! ```
//!
//! ## Supported formats
//!
//! The library supports the following formats:
//!
//! * Full Image4 containers (IMG4) via the [`ImageRef`] and [`Image`] types;
//! * Image4 payload files (IM4P) via the [`PayloadRef`] and [`Payload`] types;
//! * Image4 manifest files (IM4M) via the [`ManifestRef`] and [`Manifest`] types.
//!
//! Additionally the new payload properties format (PAYP) is supported via the [`PayloadPropsRef`]
//! and [`PayloadProps`] types, though it likely isn't stored as a separate file.
//!
//! # Features
//!
//! * `alloc`: enables owned versions of reference types, [`alloc`] support in [`der`] and decoding
//! of Image4 properties.
//! * `std`: implements [`std::error::Error`] trait on provided errors, enables [`std`] support in
//! the [`der`] crate.
//! * `payload`: enables the [`payload`] module.
//! * `manifest`: enables the [`manifest`] module.
//! * `restore_info`: enables the [`restore_info`] module.
//! * `image`: enables the [`image`] module, requires `payload`, `manifest` and `restore_info`
//! features.
//! * `serde`: implements (de)serialization for Image4 property lists. Useful for translating
//! payload properties and manifest bodies to other formats and vice versa. Disabled by default.
//! * `property`: adds support for Image4 property lists (enables the [`property`] module). These
//! are an implementation detail of the Image4 format and usually shouldn't be used directly.
//!
//! All features are enabled by default.
//!
//! [`Decode`]: der::Decode
//! [`Encode`]: der::Encode
//! [`PayloadPropsRef`]: payload::PayloadPropsRef
//! [`PayloadProps`]: payload::PayloadProps

#![no_std]
#![cfg_attr(docsrs, feature(doc_auto_cfg))]
#![warn(missing_docs)]
#![warn(missing_debug_implementations)]
#![warn(rust_2018_idioms)]

#[cfg(any(
    all(
        feature = "alloc",
        any(feature = "manifest", feature = "payload", feature = "property")
    ),
    test
))]
extern crate alloc;
#[cfg(any(feature = "std", test))]
extern crate std;

#[cfg(feature = "image")]
pub mod image;
#[cfg(feature = "manifest")]
pub mod manifest;
#[cfg(feature = "payload")]
pub mod payload;
#[cfg(feature = "property")]
pub mod property;
#[cfg(feature = "restore_info")]
pub mod restore_info;
mod tag;
mod util;

pub use der::{self, Result};
#[cfg(all(feature = "image", feature = "alloc"))]
pub use image::Image;
#[cfg(feature = "image")]
pub use image::ImageRef;
#[cfg(all(feature = "manifest", feature = "alloc"))]
pub use manifest::Manifest;
#[cfg(feature = "manifest")]
pub use manifest::ManifestRef;
#[cfg(all(feature = "payload", feature = "alloc"))]
pub use payload::Payload;
#[cfg(feature = "payload")]
pub use payload::PayloadRef;
#[cfg(feature = "restore_info")]
pub use restore_info::RestoreInfo;
pub use tag::Tag;
