//! Provides error types for the crate.

use super::{DigestAlgo, SigningAlgo};
use std::fmt;

/// An error returned by the [`FromStr`] implementation on [`DigestAlgo`] when the algorithm's name
/// isn't recognized.
///
/// [`FromStr`]: core::str::FromStr
#[derive(Clone, Debug)]
pub struct UnknownDigest(pub Box<str>);

impl std::error::Error for UnknownDigest {}

impl fmt::Display for UnknownDigest {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.write_str("unsupported digest name \"")?;
        fmt::Display::fmt(&self.0.escape_debug(), f)?;
        f.write_str("\"")
    }
}

/// An error returned when either [`SigningKey`] or [`VerifyingKey`] is constructed with unsupported
/// signing and digest algorithm pair.
///
/// [`SigningKey`]: crate::keys::SigningKey
/// [`VerifyingKey`]: crate::keys::VerifyingKey
#[derive(Clone, Debug)]
pub struct UnsupportedDigest {
    /// The signing algorithm.
    pub signing_algo: SigningAlgo,
    /// The digest algorithm.
    pub digest_algo: DigestAlgo,
}

impl fmt::Display for UnsupportedDigest {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(
            f,
            "a combination of {} signature algorithm and {} digest algorithm isn't supported",
            self.signing_algo, self.digest_algo
        )
    }
}

impl std::error::Error for UnsupportedDigest {}

/// An error returned when either a public or a private key couldn't be decoded.
#[derive(Clone, Debug)]
pub enum DecodeError {
    /// A DER decoding error.
    Asn1(der::Error),
    /// An RSA key couldn't be decoded.
    Pkcs1(rsa::pkcs1::Error),
    /// A PKCS#8 key couldn't be decoded.
    Pkcs8(pkcs8::Error),
    /// An ECDSA key couldn't be decoded.
    Sec1(sec1::Error),
    /// Unknown PEM label.
    UnexpectedLabel(String),
}

impl From<der::Error> for DecodeError {
    fn from(value: der::Error) -> Self {
        Self::Asn1(value)
    }
}

impl From<rsa::pkcs1::Error> for DecodeError {
    fn from(value: rsa::pkcs1::Error) -> Self {
        Self::Pkcs1(value)
    }
}

impl From<pkcs8::Error> for DecodeError {
    fn from(value: pkcs8::Error) -> Self {
        Self::Pkcs8(value)
    }
}

impl From<sec1::Error> for DecodeError {
    fn from(value: sec1::Error) -> Self {
        Self::Sec1(value)
    }
}

impl fmt::Display for DecodeError {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            DecodeError::Asn1(e) => write!(f, "failed to decode PEM file: {e}"),
            DecodeError::Pkcs1(e) => fmt::Display::fmt(e, f),
            DecodeError::Pkcs8(e) => fmt::Display::fmt(e, f),
            DecodeError::Sec1(e) => fmt::Display::fmt(e, f),
            DecodeError::UnexpectedLabel(label) => write!(f, "unexpected PEM label {label}"),
        }
    }
}

impl std::error::Error for DecodeError {
    fn source(&self) -> Option<&(dyn std::error::Error + 'static)> {
        match self {
            DecodeError::Asn1(e) => Some(e),
            DecodeError::Pkcs1(e) => Some(e),
            DecodeError::Pkcs8(e) => Some(e),
            DecodeError::Sec1(e) => Some(e),
            DecodeError::UnexpectedLabel(_) => None,
        }
    }
}
