use crate::{comp::ComprAlgo, util};
use clap::{Args, Parser, Subcommand};
use der::referenced::OwnedToRef;
use image4::{der::Encode, payload::ComprInfo, PayloadRef, Tag};
use std::{
    fs::File,
    io::{Read, Write},
    path::PathBuf,
};

#[derive(Args, Debug)]
pub struct PayloadCreationOptions {
    /// Sets the builder string for the created payload.
    #[arg(long = "builder-string", short = 'b')]
    pub builder_string: Option<String>,

    /// Adds an arbitrary compression info blob to the generated payload.
    ///
    /// This shouldn't be used directly unless you know what you're doing. The option may be useful
    /// when an unsupported compression algorithm is added.
    #[arg(long = "raw-compr-info", value_parser = util::parse_raw_compr_info)]
    pub raw_compr_info: Option<ComprInfo>,

    /// Compresses the payload contents using the specified algorithm.
    ///
    /// Currently only LZFSE is supported.
    #[arg(long = "compress", value_name = "ALGO", requires = "tag")]
    pub compr_algo: Option<ComprAlgo>,

    /// Adds payload properties described by the specified Property List file to the Image4 payload.
    #[arg(long = "payload-props", value_name = "PATH", requires = "tag")]
    pub props: Option<PathBuf>,
}

#[derive(Parser, Debug)]
/// Creates an Image4 payload file from raw payload contents.
pub struct CreateArgs {
    /// A 4CC tag for the created payload file.
    #[arg(long, short, value_name = "TAG")]
    tag: Tag,

    /// Forces the tool to continue in cases when the user would be asked for confirmation.
    #[arg(long, short)]
    force: bool,

    #[command(flatten)]
    common: PayloadCreationOptions,

    /// A path to the input file containing the raw contents of the newly created payload.
    #[arg(value_name = "INPUT")]
    input: PathBuf,

    /// A path where the newly created Image4 payload (IM4P) file will be written.
    ///
    /// When this option is missing, the file will be created adjacent to the input file. The name
    /// will be the same as that of the input file with the extension set to ".im4p".
    #[arg(value_name = "OUTPUT")]
    output: Option<PathBuf>,
}

#[derive(Subcommand, Debug)]
#[command(name = "payload")]
/// Transforms Image4 payloads (IM4P files).
pub enum PayloadCommand {
    Create(CreateArgs),
}

pub fn do_create(args: &CreateArgs) -> anyhow::Result<()> {
    let mut input_file = File::open(&args.input)?;

    let output_path =
        util::construct_output_path(args.output.as_deref(), &args.input, "im4p", &["bin"]);

    info!(
        "Creating a payload file with contents of \'{}\' at \'{}\'.",
        args.input.display(),
        output_path.display()
    );

    let mut input_data = Vec::new();
    input_file.read_to_end(&mut input_data)?;

    let compr_info = util::maybe_compress_payload(args.common.compr_algo, &mut input_data)?;

    let builder_string = args
        .common
        .builder_string
        .as_deref()
        .unwrap_or(crate::DEFAULT_BUILDER_STRING);
    let mut payload = PayloadRef::new(args.tag, builder_string, &input_data)?;

    if let Some(raw_compr_info) = &args.common.raw_compr_info {
        info!(
            "Adding raw compression info (algo is {}, uncompressed size is {}).",
            raw_compr_info.algo, raw_compr_info.uncompressed_size
        );

        payload.set_compr_info(Some(raw_compr_info.clone()));
    } else if let Some(compr_info) = compr_info {
        payload.set_compr_info(Some(compr_info));
    }

    let props = util::maybe_load_payload_props(args.common.props.as_ref())?;
    payload.set_props(props.as_ref().map(OwnedToRef::owned_to_ref));

    info!("4CC tag: {}", args.tag);
    info!("Builder string: {builder_string}");

    let mut payload_data = Vec::new();
    payload.encode_to_vec(&mut payload_data)?;

    let mut output_file = util::try_open_output(&output_path, args.force)?;
    output_file.write_all(&payload_data)?;

    Ok(())
}
